module FailList (FailList (),failnull,failures,failhead,failmaybe,
                 failedmap,faillist) where
import GList

data FailList a b = FailList [a] [b] -- list of failures, list of successes
  deriving Show

listify :: FailList a b -> [b]
listify (FailList _ xs) = xs

failnull :: FailList a b -> Bool
failnull = null.listify

failhead :: FailList a b -> b
failhead = head.listify

failures :: FailList a b -> [a]
failures (FailList xs _) = xs

failmaybe :: [a] -> [b] -> FailList a b
failmaybe xs [] = FailList xs []
failmaybe _ ys = FailList [] ys

faillist xs ys = FailList xs ys

failedmap :: (a -> b) -> (FailList a c -> FailList b c)
failedmap f (FailList xs ys) = FailList (map f xs) ys

instance Monad (FailList a) where
   return y = FailList [] [y]
   (FailList xs [])     >>= f = FailList xs []
   (FailList xs (y:ys)) >>= f = f y `gcat` ((FailList xs ys) >>= f)
   fail s = FailList [] []

   -- Note that the Monad instance is only valid if the order of the
   -- list of failures is not considered important.

   -- return y >>= f = FailList [] [y] >>= f
   --                = f y `gcat` (FailList [] [] >>= f)
   --                = f y `gcat` FailList [] []
   --                = f y

   -- FailList xs [] >>= return = FailList xs []
   -- FailList xs (y:ys) >>= return = return y `gcat` 
   --                                      (FailList xs ys >>= return)
   --                               = FailList [] [y] `gcat` FailList xs ys
   --                               = FailList xs (y:ys)

   -- FailList xs [] >>= (\x -> k x >>= h) = FailList xs []
   --                                        = FailList xs [] >>= h
   --                                        = (FailList xs [] >>= k) >>= h
   -- FailList xs (y:ys) >>= (\x -> k x >>= h) 
   --            = k y >>= h `gcat` FailList xs ys >>= (\x -> k x >>= h)
   --            = k y >>= h `gcat` (FailList xs ys >>= k) >>= h

   -- Lemma: (a `gcat` b) >>= f = a >>= f `gcat` b >>= f

   -- (FailList xs [] `gcat` FailList xs' []) >>= f
   --     = FailList (xs++xs') [] >>= f
   --     = FailList (xs++xs') []
   --     = FailList xs [] `gcat` FailList xs' []
   --     = FailList xs [] >>= f `gcat` FailList xs' [] >>= f
   
   -- (FailList xs [] `gcat` FailList xs' (y':ys')) >>= f
   --     = FailList (xs++xs') (y':ys') >>= f
   --     = f y' `gcat` FailList (xs++xs') ys' >> = f
   --     = f y' `gcat` (FailList xs [] `gcat` FailList xs' ys') >>= f
   --     = f y' `gcat` (FailList xs [] >>= f `gcat` FailList xs' ys' >>= f)
   --     = FailList xs [] >>= f `gcat` (f y' `gcat` FailList xs' ys' >>= f)
   --     = FailList xs [] >>= f `gcat` FailList xs' (y':ys') >>= f
   
   -- (FailList xs (y:ys) `gcat` FailList xs' ys') >>= f
   --     = FailList (xs++xs') (y:ys++ys') >>= f
   --     = f y `gcat` FailList (xs++xs') (ys++ys') >>= f
   --     = f y `gcat` (FailList xs ys >>= f `gcat` FailList xs' ys' >>= f)
   --     = (f y `gcat` FailList xs ys >>= f) `gcat` FailList xs' ys' >>= f
   --     = FailList xs (y:ys) >>= f `gcat` FailList xs' ys' >>= f
   


instance Functor (FailList a) where
   fmap f (FailList xs ys) = FailList xs (map f ys)

instance MonadPlus (FailList a) where
   mzero = FailList [] []
   mplus (FailList xs ys) (FailList xs' ys') = FailList (xs++xs') (ys++ys')

instance GList (FailList a) where
   gfilter p (FailList xs ys) = FailList xs (filter p ys)
   gsequence (FailList xs cs) = do ys <- sequence cs
                                   return (FailList xs ys)

   -- override gconcat to provide a decently performing implementation
   -- this makes a *BIG* difference!
   gconcat (FailList xs yss) = FailList (xs++xs') ys
       where FailList xs' ys = gconcat' yss

   -- gconcat (FailList xs []) = FailList xs []
   -- gconcat (FailList xs (FailList xs' ys':ys)) 
   --    = FailList (xs'++xs'') (ys'++ys'')
   --       where FailList xs'' ys'' = gconcat (FailList xs ys)
